<?php

/**
 * 2009-08-27 messageUnique.php
 *
 * @author Mariusz Chwalba <koder@mediasystems.pl>
 * @copyright 2009 Media Systems
 * @package cashbill
 *
 * Obsługa unikalności identyfikatora.
 * 
 */
/**
 * @page errorHandling Zabezpieczenia przed błędami
 * 
 * @section uniqueId Unikalność identyfikatora
 * 
 * Transmisja danych w publicznej sieci internetowej obarczona jest ryzykiem
 * wystąpienia błędu niezależnego od partnera i systemu Cashbill. Z tego względu
 * protokół transmisji danych wyposażony został w mechanizmy automatycznej
 * kontroli i korekcji błędu.
 * 
 * Protokół transmisji dopuszcza wielokrotne przesyłanie tej samej wiadomości
 * jeżeli tylko istnieje jakaś wątpliwość co do poprawności jej przetworzenia.
 * Aby uniknąć wielokrotnego wysłania tej samej wiadomości do klienta każda
 * wiadomość \e musi być wyposażona w unikalny identyfikator.
 * 
 * Jeżeli w transmisji pojawi się ponownie już obsłużona wiadomość (wiadomość
 * o identyfikatorze który został już poprawnie przetworzony), protokół transmisji
 * przewiduje odpowiedź "ACCEPTED" oraz zignorowanie wiadomości. 
 * 
 * @subsection uniqueIdOut Identyfikatory wiadomości wychodzących
 * 
 * Jeżeli partner pominie identyfikator na dowolnym etapie wysyłania wiadomości
 * bramka postara się utworzyć odpowiedni identyfikator automatycznie. Kryteria
 * tworzenia identyfikatora zależne są od metody tworzenia wiadomości wychodzącej:
 * 
 * -# Dla wiadomości utworzonej metodą reply z cMessageSMSIn identyfikator powstaje
 *    na podstawie id wiadomości inicjującej oraz numeru odpowiedzi.
 * -# Dla wiadomości utworzonej bezpośrednio z cMessageSMSOut, w przypadku braku
 *    identyfikatora tworzony jest on na podstawie numeru docelowego, treści oraz
 *    czasu dostarczenia.
 *    
 * @subsection uniqueIdIn Identyfikatory wiadomości przychodzących
 * 
 * Każda wiadomość dostarczana przez system Cashbill do serwera partnera posiada
 * unikalny identyfikator. Ta sama wiadomość może zostać dostarczona wielokrotnie,
 * jeżeli system Cashbill uzna, iż istnieje ryzyko błędu transmisji. W szczególności,
 * ponowne dostarczenie wiadomości nastąpi gdy serwer partnera nie odpowiedział zgodnie
 * z protokołem SOAP. Ta sytuacja może mieć następujące przyczyny:
 * 
 * -# Występują problemy z transmisją sieciową.
 * -# Wystąpiła awaria serwera partnera.
 * -# Oprogramowanie partnera zawiera błędy.
 * -# Serwer partnera zbyt długo przetwarza wiadomość (dłużej, niż timeout protokołu). 
 * -# Serwer partnera odpowiedziął wyjątkiem (SoapFault)
 * 
 * Ponowne dostarczanie wiadomości zapewnia, że każda wiadomość wysłana przez klienta
 * zostanie dostarzcona do partnera.
 *
 * W przypadku większości usług opartych o SMS bezpośredni ponowne przetworzenie
 * wiadomości od klienta jako nowej jest nieporządane - skutkuje przyznaniem wielu
 * kodów dostępu, zaliczeniem większej niż faktyczna liczby głosów w sondzie, itp.
 * 
 * Z tego względu zalecamy by dla wszystkich usług weryfikować unikalność identyfikatora.
 * W przypadku stwierdzenia ponownego przetworzenia tej samej wiadomości handler
 * partnera może zarządać zignorowania wiadomości wyjątkiem:   
 * 
 * @code
 * 
 * throw new cGatewayException( 'Wiadomość została już obsłużona', cGatewayException::IGNORE_MESSAGE );
 * 
 * @endcode
 * 
 * Partner może wykorzystać dowolny opis wyjątku do celów diagnostycznych.
 * 
 * @subsection uniqueIdInTemplate Gotowe rozwiązania
 * 
 * Aby uprościć implementację usług opartych o SMS bezpośredni w niniejszym pakiecie
 * zawarto gotowe rozwiązania umożliwiające przechwycenie wielokrotnej transmisji
 * tej samej wiadomości.
 * 
 * Uruchomienie gotowej obsługi wielokrotnej transmisji tej samej wiadomości
 * Wystarczy \e przed właściwym handlerem obsługi wiadomości umieścić kod: 
 *
 * @code
 * $smsGateway->addSMSHandler( new cSMSUniqeIdMySqlCheck( <parametry połączenia> [,nazwa tabeli] ) );
 * @endcode
 * 
 * A po zakończeniu obsługi wiadomości:
 * 
 * @code
 * $smsGateway->addSMSHandler( new cSMSUniqeIdMySqlDone( <parametry połączenia> [,nazwa tabeli] ) );
 * @endcode
 * 
 * W obu przypadkach klasy oczekują przekazania tego samego połączenia do bazy danych.
 * Połączenie jest deklarowane z wykorzystaniem klasy cMySqlConnection.
 * 
 * Uzupełniony kod obsługi wiadomości wygląda tak:
 * 
 * @code
 * $mysql = new cMySqlConnection( <host>, <login>, <hasło>, <baza danych> );
 * $smsGateway = new cGateway( <klucz tajny>, <prefiks> );
 * 
 * $smsGateway->addSMSHandler( new cSMSUniqeIdMysqlCheck( $mysql ));
 * $smsGateway->addSMSHandler( new cMojaObslugaSms() );
 * $smsGateway->addSMSHandler( new cSMSUniqeIdMysqlDone ( $mysql ));
 *  
 * $smsGateway->handleInput(); 
 * @endcode
 * 
 * 
 * Gotowe rozwiązania zawarte w plikach template_*.php zawierają te linie, jeżeli tylko
 * jest konieczna do poprawnego funkcjonowania danego schematu usługi. 
 * 
 */


require_once 'messageSqlBased.php';

/**
 * Handler wiadomości SMS zapewniający poprawną obsługę błędów.
 * Sprawdza czy wiadomość została wcześniej przysłana, czy proces obsługi został
 * zakończony poprawnie oraz czy nadal trwa. Zapewnia zgodną z protokołem obsługę
 * wielokrotnie przesyłanych wiadomości.
 * 
 */
class cSMSUniqeIdMySqlCheck extends cSMSMySqlBased implements iSMSHandler
{
	public function handle( cMessageSMSIn $sms )
	{
		static $realPid;
        if ( !$realPid ) $realPid = uniqid('smsPid.');
		
		$id = mysql_escape_string( $sms->id );
        
		// pozyskanie locku procesowego MySql
		// lock jest zwalniany przez MySql automatycznie w momencie zakończenia połączenia
        $lock = mysql_fetch_object( mysql_query( "SELECT GET_LOCK('{$realPid}',0) AS locked", $this->connection ) );
        if ( !$lock->locked )
        {
        	throw new Exception("MySQL - cannot obtain unique lock");
        }
        
		$res  = mysql_fetch_object( mysql_query( "SELECT `id`,`pid`,`done` FROM `{$this->table}` WHERE `id`='{$id}'", $this->connection ) );
		
		if ( $res )
		{
			// sprawdzenie czy wiadomość już została obsłużone
			if ( $res->done )
			{
				throw new cGatewayException( "Message is already processed", cGatewayException::IGNORE_MESSAGE );
			}
			
			if ( $res->pid != $realPid )
			{
				/*
				 * Sprawdzenie czy proces aktualnie działający obsługuje wiadomość.
				 * Jeżeli obsługuje, to nie powinniśmy ponownie przyjmować wiadomości, ale prosić o przysłanie ponowne później.
				 */
				$free = mysql_fetch_object( mysql_query( "SELECT IS_FREE_LOCK('{$res->pid}') AS free" ) );
				if ( !$free->free )
				{
					throw new cGatewayException( "Message is being processed right now", cGatewayException::RETRY_LATER );
				}
			}
		}
		// zarejestrowanie tego procesu jako obsługującego wiadomość
		@mysql_query( "REPLACE INTO `{$this->table}` SET `id`='{$id}', `pid`='{$realPid}'");
	}
}

/**
 * Handler wiadomości SMS obsługi błędów, zaznaczający zakończenie prac nad wiadomością.
 * Powinien być dodany do bramki jako ostatni. Jego wywołanie zaznaczy wiadomość jako obsłużoną.
 *
 */
class cSMSUniqeIdMySqlDone extends cSMSMySqlBased implements iSMSHandler
{
	public function handle( cMessageSMSIn $sms )
	{
		$id = mysql_escape_string( $sms->id ); 
		@mysql_query( "UPDATE `{$this->table}` SET `done`=1 WHERE `id`='{$id}'", $this->connection );
	}
}


?>